/*  Copyright 2023 Dieter Holzhäuser  E-Mail: dieter.holzhauser@gmail.com
  
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published serbuf
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "mtRTOS2407.h"

#define TIMERINTHANDLER  ISR(TIMER2_OVF_vect)
#define TIMERRESTART     TCNT2 = 6;
void inittimer () {
  TCCR2A   = 0x00;      // Bit       1 0
                        //           0 0  Normal Mode
  TCCR2B   = 0x04 ;     // Bit   3 2 1 0
                        //         1 0 0  Prescaler 1/64
                        //       0        Normal Mode
  TCNT2 = 6;            //Start Timer2, 256-6=250 ticks to overflow
  TIMSK2 |= (1 <<  TOIE2);    //Interrupt on overflow enable
}

/////////////////////////////////////////////////////////////////
/////////////////////////Multitasking Kernel  ///////////////////
      //Task state elements (masks)
#define  WAITING           7    //00000111
#define  SEMABIT           8    //00001000
#define  TIMERBIT         16    //00010000
#define  PINBIT           32    //00100000

///////////////////
long clock = 0;            //ms since start of program
long sclock = 0;           //s since start of programm
long stime = 0;            //counter of seconds
unsigned int idle;         //minimal number of runner idle cycles per ms
                           //during last IDLEPERIOD  (like max CPU load)

unsigned char runtask = NIL;    //running task

unsigned int runcycs = 0;      //idle data
unsigned int timercycs = 0;    
unsigned int idlemin = 0xFFFF;  

struct {                      //Array of task data structures
  unsigned char next;
  unsigned char state;
  void (*mainpfunc)(void) ;
  long timer;
  void (*altpfunc)(void) ;
  unsigned char sema;
  unsigned char edge;
  unsigned char pinold;
  unsigned char pinnr;
} tasklist[MAXTASKS] ;              

unsigned char semalist[MAXSEMAS];   //Array of sema data
           // 2 linked lists are embedded in tasklist by component next
unsigned char readytop = NIL;  //linked list readylist top index
unsigned char waittop = NIL;   //linked list waitlist  top index

///////////////linked list functions (non puplic)  ////////////////
                                 //append task to linked list ready/wait
void append (unsigned char task, unsigned char * ptop){
  unsigned char i, prec = 0;
  if (ptop == &readytop) tasklist[task].state = READY;
  if (*ptop == NIL)
    *ptop = task;                       //new
  else {
    for (i = *ptop ; i != NIL; i = tasklist[i].next)  prec = i;
    tasklist[prec].next = task;         //append
  }
  tasklist[task].next = NIL ;
}
                                 //delete task from from linked list ready/wait
void del (unsigned char task, unsigned char* ptop ) {
  unsigned char i , prec = 0;
  for (i = *ptop; i != NIL; i = tasklist[i].next) {
    if ( i == task) {
      if (i == *ptop)
        *ptop = tasklist[i].next;                   //from top position
      else tasklist[prec].next = tasklist[i].next;  //other position
    }
    prec = i;
  }
}
////////////////////////
TIMERINTHANDLER  {
  TIMERRESTART
  clock++;
  if (runcycs < idlemin ) idlemin = runcycs;     //Manage idle
  runcycs = 0;
  if (++timercycs >= ONESECOND ) {
    stime++;
    sclock++;
    timercycs = 0;
    idle = idlemin; 
    idlemin = 0xFFFF;
  }
  unsigned char i ;
  for (i = 0; i != MAXTASKS; i++ ) {
    if (tasklist[i].state & PINBIT ) {            //detect WAITING_B input pins
      unsigned char pinact =  (unsigned char) RPIN ( tasklist[i].pinnr );
      if (  pinact !=   tasklist[i].pinold ) {
        tasklist[i].pinold   =  pinact;
        if ((pinact  &&  tasklist[i].edge) ||         //rising edge or
              (!pinact &&  ! tasklist[i].edge)) {     //falling edge
          tasklist[i].mainpfunc = tasklist[i].altpfunc;
          append (i , &readytop);                        //set event
          continue;
        }
      }
    }
    if (tasklist[i].state & TIMERBIT ) {            //Manage task timer
      if (--tasklist[i].timer == 0) {
        if (tasklist[i].state & SEMABIT)  del (i, &waittop );
        append (i , &readytop);                     //set event
      }
    }
  }
}

//////////////////////// public functions ///////////////
void runner (void) {     //"motor" of program,  call in setup, avoid loop() 
  while (1) {
    if ( readytop != NIL  ) {
      DI
      runtask = readytop;
      readytop = tasklist[runtask].next;
      tasklist[runtask].state = RUNNING;
      EI
      tasklist[runtask].mainpfunc ();            //call task function
      DI
      if (tasklist[runtask].state == RUNNING) {  //task function returned
        tasklist[runtask].state = TERMINATED;    //without mt function call
        del (runtask, &waittop);
      }
      EI
    }
    else runcycs++;                   //count idle cycles
  }
}

void start (unsigned char task, void (*pfunc) (void) ) {
  DI
  if (tasklist[task].state == TERMINATED)   {
    tasklist[task].mainpfunc = pfunc;
    append (task , &readytop);
  }
  EI
}

void stop (unsigned char task, void (*pfunc) (void) ) {   
  if (tasklist[task].state != TERMINATED) {
    if ( pfunc ) pfunc ();     //function to create a defined state or zero
    DI
    del ( task, &readytop);
    del (task, &waittop);
    tasklist [task].state = TERMINATED;
    EI
  }
}

//////Info
long getclock (void){
  return clock;
}
long getsclock (void) {
  return sclock;
}
long getstime (void){
  return stime;
}
void setstime (long t) {
  stime = t;
}
long getidle(void){
  return idle;
}
long gettimer (unsigned char task   ) {
  unsigned long t;
  t = tasklist[task].timer;
  return t;
}
unsigned char gettaskstate (unsigned char task) {
  unsigned char n;
  n = tasklist[task].state;
  return n;
}
unsigned char gettasksema (unsigned char task) {
  unsigned char n;
  if (tasklist[task].state == WAITING+SEMABIT)
    n = tasklist[task].sema;
  else n = NIL;
  return n;
}
unsigned char getruntask (void) {
  return runtask;
}
unsigned char issema (unsigned char sema ) {
  unsigned char n;
  n = semalist [sema];
  if (n) return 1; else return 0;
}

/////mt functions
void mtcoop ( void (*pfunc) (void)   ) {
  DI
  if (tasklist[runtask].state == RUNNING) {
    tasklist[runtask].mainpfunc = pfunc;
    append (runtask, &readytop);
  }
  EI
}
void mtsema (unsigned char sema , void (*pfunc) (void) ) {
  DI
  if (tasklist[runtask].state >= RUNNING) {
    if ( semalist[sema] > 0 ) {
      semalist[sema] = 0;
      tasklist[runtask].mainpfunc = pfunc;
      append (runtask , &readytop);     //state READY, like mtcoop
    }
    else {
      tasklist[runtask].sema = sema;
      tasklist[runtask].altpfunc = pfunc;
      tasklist[runtask].state |= WAITING + SEMABIT;
      tasklist[runtask].state &= ~PINBIT;
      append (runtask , &waittop);
    }
  }
  EI
}
void mtdelay ( long t , void (*pfunc) (void)  )  {
  DI
  if (t == 0) mtcoop (pfunc);
  else {
    if ( tasklist[runtask].state >= RUNNING) {
      tasklist[runtask].timer = t;
      tasklist[runtask].mainpfunc = pfunc;
      tasklist[runtask].state |= WAITING + TIMERBIT;
    }
  }
  EI
}
void mtpin ( unsigned char pinnr, unsigned char edge, void (*pfunc) (void) ) {
  DI
  if ((tasklist[runtask].state >= RUNNING)) {
    tasklist[runtask].pinnr  = pinnr;
    tasklist[runtask].edge  = edge;
    tasklist[runtask].pinold  =  (unsigned char) RPIN (pinnr);
    tasklist[runtask].altpfunc = pfunc;
    tasklist[runtask].state |= WAITING + PINBIT;
    tasklist[runtask].state &= ~SEMABIT;
  }
  EI
}

/////Sema functions
void signalisr (unsigned char sema ) {
  unsigned char i , prec = 0;
  for (i = waittop; i != NIL; i = tasklist[i].next) {
    if  (tasklist[i].sema == sema) {         //take out of waitlist
      if (i == waittop)
          waittop = tasklist[i].next;                   // top position
      else tasklist[prec].next = tasklist[i].next;      // other position
      tasklist[i].mainpfunc = tasklist[i].altpfunc;
      append (i , &readytop);
      return;
    }
    prec = i;
  }
  semalist [sema] = 1;
}

void signal (unsigned char sema ) {
  DI
  signalisr (sema);
  EI
}

void clearsema (unsigned char sema ) {
  semalist [sema] = 0;
}

///////////////////////Setup 
void initmt () {           // call in Arduino setup ()
  unsigned char i ;
  for ( i = 0; i < MAXTASKS; i++) tasklist[i].state = TERMINATED;
  for ( i = 0; i < MAXSEMAS; i++) semalist[i] = 0; 
  inittimer ();
  EI
}
///////////////////////////////////////////////////////////////
//////////////////////////////Serial///////////////////////////
char deci;
char minus;
long num;
char ch; 
//////////////// 
char inum;
long lnum [MAXLNUM];
char id;           
////////////////
char prcode;   //processing code 
//errors       F20 illegal id character    
//             F21 illegal character  
//             F22 type long out of range 
void prt_error (int n) {
  PR ("F");           
  PRln (n);
}
///////////////////Task SUI    serial user input
void sui_home ();
void sui_next ();

void sui_clear (){                //clear buffer 
  ch = Serial.read ();  
  if (ch == -1) mtcoop (sui_home); else mtcoop (sui_clear);   
}

void sui_num () {
  if (prcode >= 20) {
    prt_error (prcode); 
    mtcoop (sui_clear);
    return;
  }
  ch = Serial.read ();  
  if (ch == CR) ch = LF;   
  switch (prcode) {     //there are different types of evaluation depending on prcode 
    case 0:             //check first numeric character 
      if (ch == ' ') 
        ;               //ignore leading blanks
      else      if (ch == '-') {  
                  if (minus) prcode = 21;      //illegal character   
                  minus = 1;
                }
                else if (ch == '#')      
                       prcode =  2;
                     else if (ch >= '0' && ch <= '9') {                          
                            num = ch - '0';     //initialize transformation character to num
                            prcode = 1;   
                          } 
                          else prcode = 21;     //illegal character                       
      break;
    case 1:                             //check next numeric character 
      if (ch >= '0' && ch <= '9') {     //transformation 
        if (deci < MAXDECI+1) {
          if (num <= 214748364) {         //least significant decimal
            num = num * 10;
            num = num + ch - '0';
            if ( num < 0 ) prcode = 22;      //out of range
          }  
          else prcode = 22;                  //out of range
          if (deci) deci++;  
        }              
      }
      else if (ch == '.')  {                //decimal (special use)  
             if (deci) prcode = 21; else deci = 1;   // illegal character                        
           }
           else if (ch == ' ' || ch == LF) {
                  if (minus) num = -num;   
                  if (deci) while (deci++ < MAXDECI+1) num = num * 10; 
                  lnum [inum++] = num; 
                  goto l1;
                }
                else prcode = 21;            // illegal character 
        break;
      case 2:   
        lnum [inum++] = NONB; 
        if (ch == LF || ch == ' ') {
l1:       if (ch == LF || inum == MAXLNUM) {
            signal (INPSEM);   
            mtcoop (sui_clear);
          }
          else mtcoop (sui_next);
          return;
        }
        else prcode = 21;                    //illegal character
        break;
  }
  mtcoop (sui_num); 
}

void sui_next () {
  prcode = 0;  
  minus = 0; 
  deci  = 0;
  mtcoop ( sui_num );  
}

void sui_id (){             //check for id 
  ch = Serial.read ();  
  if (ch == CR) ch = LF;   
  if (ch == ' ') {
    if (id == 0) mtcoop (sui_id); else mtcoop (sui_next);  
    return;
  }
  else if (ch == LF)
         signal (INPSEM);
       else {
         if (( ch >= 'a' && ch <= 'z') || ( ch >= 'A' && ch <= 'Z')){   
              if (id == 0)  {
                id = ch;
                mtcoop (sui_id);
                return;
              }
         }      
         prt_error (20);             //illegal id character
       }
  mtcoop (sui_clear); 
}

void sui_home (){                    //detect serial input
  if ( Serial.available() ) {       
    id = 0; 
    inum = 0;
    mtcoop (sui_id);                 
  }    
  else mtdelay (100, sui_home);  
}
            
void sui1 () {                   
  mtcoop (sui_home);
}

void initserial (long baud) {
  Serial.begin (baud); 
  start (SUI, sui1); 
}

///////////////////////EEMEM//////////////////
//////////////read
void eeread (unsigned int eeadr, unsigned char size, void * pvar ) {
  unsigned char n;
  for ( n = 0; n < size; n++) {
    DI
    eeadr = eeadr + n;
    EEARL = eeadr % 256;    //Set up address register
    EEARH = eeadr / 256;
    EECR |= (1<<EERE);      //Start eeprom read by writing EERE
    ((unsigned char *) pvar) [n] = EEDR;   //data from data register
    EI
  }
  signal (EESEM);
}
///////////write
void (*eeretf)(void);      //save task function to continue task
unsigned char *spvar;      //save pointer to first byte of var
unsigned char  ssize;      //save size of var
unsigned int   seeadr;     //save EEMEM adress
unsigned char  een;        //counter

void ee1 () {                   
  if ( !(EECR & (1<<EEPE) ) ) {  // completion of previous write
    if (een == ssize ) {         //end of eewrite
      signal (EESEM);
      mtcoop (eeretf);
      return;                
    }
    DI
    EECR = (0<<EEPM1)|(0<<EEPM0);   //Set Programming mode
    seeadr = seeadr + een;          //Set up address registers
    EEARL = seeadr % 256;
    EEARH = seeadr / 256;
    EEDR =  spvar[een];          //Set up data register
    EECR |= (1<<EEMPE);          //Write logical one to EEMPE
    EECR |= (1<<EEPE);           //Start eeprom write by setting EEPE
    EI
    een++;
  }
  mtdelay (1, ee1);
}
                       //init subtask eewrite
void eewrite (void * pvar, unsigned char size, unsigned int eeadr, void (*retf)(void) ) {
  eeretf = retf;
  spvar = (unsigned char *) pvar;
  seeadr = eeadr;
  ssize = size;
  een = 0;
  mtcoop (ee1);
}






