/*  Copyright 2023 Dieter Holzhäuser  E-Mail: dieter.holzhauser@gmail.com
  
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published serbuf
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MTRTOS2407_H_INCLUDED
#define MTRTOS2407_H_INCLUDED

#include <Arduino.h>

//MTRTOS  Multitasking Realtime Operating System
//realtime compatible serial input
//realtime compatible use of EEMEM

/////////////////Serial input library interface 
//Semaphore  INPSEM   ->  serial input has taken place
//extern char id      ->  ID:   0 (LF), 'a'to'z' , 'A'to'Z'
//extern char inum;   ->  number of numeric data, if inum == 0 means id only
//extern long lnum [MAXNUM]  ->  array of numeric data 

//     Serial input format 
//  LF  or  id LF  or  id num (num) (num) (num) LF    (blank(s) as separator)
//     Format of num (digits 123 as example)
//  #  or  123  or  -123  or  1.23  or  -1.23

#define RPIN    digitalRead
#define WPIN    digitalWrite
#define PR      Serial.print
#define PRln    Serial.println
#define DI      SREG = SREG & ~0x80;    //Disable Interrupt
#define EI      SREG = SREG | 0x80;     //Enable Interrupt

#define MAXTASKS  20
///////////Library Tasks
#define  SUI      MAXTASKS-1    //Task   Serial User Input
                                //possible application tasks: 0 to MAXTASKS-2
#define  MAXSEMAS   20    
///////////////Library Semas  
#define  INPSEM     MAXSEMAS -1  
#define  EESEM      MAXSEMAS -2                                  
                                  //possible application semas: 0 to MAXSEMAS-3
                 
////////////////////////////////////////////////////////////
////////////////////Multitasking Kernel  ///////////////////
//call in setup:   initmt ()    as first library init
//                 start (...) 
//                 ...
//                 runner ()    as last statement, avoid function loop ()

#define  ONESECOND  1000         //ms  idle analyse period, do not change
#define  NIL        255

/////Task states, do not change
#define READY        0
#define TERMINATED   5
#define RUNNING      6
#define WAITING_S    15     //Semaphore (Sema)
#define WAITING_T    23     //Timeout
#define WAITING_B    39     //Input pin bit edge
#define WAITING_TS   31     //Both Timeout or Sema
#define WAITING_TB   55     //Both Timeout or Input pin bit edge

void runner (void);
//endless, task functions are called ("motor" of program)

void start  (unsigned char task, void (*pfunc) (void) );
//Start of task by task function pfunc

void stop (unsigned char task, void (*pfunc) (void) );
//Call function pfunc and stop task

///////Info
long getidle(void);    //minimal number of runner idle cycles per ms
                       //while last ONESECOND period (like inversed CPU load)

long getclock (void);  //time (ms) program is running

long getsclock (void); //time (s) program is running
  
long getstime (void);    //basic clock (counter of seconds) 

void setstime (long t);  //set basic clock
 
unsigned char gettaskstate (unsigned char task);

unsigned char gettasksema (unsigned char task);
                        //sema if task is WAITING_S else ŃIL

unsigned char getruntask (void);     //actual running task

long gettimer (unsigned char task);  //cycles until timeout

unsigned char issema (unsigned char sema );  //1 if sema is active else 0

//////////////mt functions
//mt functions are setting data, so runner is able to continue related 
//tasks by calling pfunc.  mt functions have to be called in each
//task function. If not task stops.

void mtcoop (void (*pfunc) (void) );
//task is continued without event

void mtsema (unsigned char sema , void (*pfunc) (void) );
//task is continued on event sema and sema is cleared

void mtdelay ( long t , void (*pfunc) (void)  );  // t ms timer
//task is continued on event timeout (t == 0)

void mtpin ( unsigned char pinnr, unsigned char edge, void (*pfunc) (void) ) ;
//task is continued on event edge of bit (HIGH or LOW) of pinnr

/////////Sema functions
void signal (unsigned char sema);
//if any task is waiting for sema (WAITING_S), task gets READY, if not, sema is set

void signalisr (unsigned char sema);
//like signal, special use in ISR

void clearsema (unsigned char sema);

////////////               
void initmt  ();      // call in setup()

////////////////////////////SERIAL/////////////////////////
#define  NONB  -2147483648        //numeric equivalent of exceptions
#define  LF    10                 //line feed
#define  CR    13 

//errors       F20 illegal id character    
//             F21 illegal character  
//             F22 out of type long range 

//Serial input library interface 
#define MAXDECI 2           // decimal digits
extern char id;             //ID
extern char inum;           
#define MAXLNUM 4    
extern long lnum [MAXLNUM];

void initserial (long baud);

/////////////////////////////////////////////
//////////////////////EEMEM/////////////////
// EEMEM adresses (eeadr) have to be defined. Compiler should not control them. 
// call eewrite as subtask,  EESEM is exclusion semaphore 

void eeread (unsigned int eeadr, unsigned char size, void * pvar );
//if call can take place during write operation, call mtsema (EESEM,...) first.

//eewrite is a subtask
void eewrite (void * pvar, unsigned char size, unsigned int eeadr, void (*retf)(void) ) ;
//if call can take place during write operation, call mtsema (EESEM,...) first.

#endif    // MTRTOS2407_H_INCLUDED
 


